/*
   mkvmerge -- utility for splicing together matroska files
   from component media subtypes

   Distributed under the GPL v2
   see the file COPYING for details
   or visit https://www.gnu.org/licenses/old-licenses/gpl-2.0.html

   helper functions for W64 data

   Written by Moritz Bunkus <mo@bunkus.online>.
*/

#include "common/common_pch.h"

#include "common/w64.h"

namespace mtx::w64 {

uint8_t const g_guid_riff[16] = {
  'r', 'i', 'f', 'f',
  0x2e, 0x91, 0xcf, 0x11, 0xa5, 0xd6, 0x28, 0xdb, 0x04, 0xc1, 0x00, 0x00
};

uint8_t const g_guid_wave[16] = {
  'w', 'a', 'v', 'e',
  0xf3, 0xac, 0xd3, 0x11, 0x8c, 0xd1, 0x00, 0xc0, 0x4f, 0x8e, 0xdb, 0x8a
};

uint8_t const g_guid_fmt[16] = {
  'f', 'm', 't', ' ',
  0xf3, 0xac, 0xd3, 0x11, 0x8c, 0xd1, 0x00, 0xc0, 0x4f, 0x8e, 0xdb, 0x8a
};

// uint8_t const g_guid_fact[16] = {
//   'f', 'a', 'c', 't',
//   0xf3, 0xac, 0xd3, 0x11, 0x8c, 0xd1, 0x00, 0xc0, 0x4f, 0x8e, 0xdb, 0x8a
// };

uint8_t const g_guid_data[16] = {
  'd', 'a', 't', 'a',
  0xf3, 0xac, 0xd3, 0x11, 0x8c, 0xd1, 0x00, 0xc0, 0x4f, 0x8e, 0xdb, 0x8a
};

// uint8_t const g_guid_summarylist[16] = {
//   0xbc, 0x94, 0x5f, 0x92,
//   0x5a, 0x52, 0xd2, 0x11, 0x86, 0xdc, 0x00, 0xc0, 0x4f, 0x8e, 0xdb, 0x8a
// };

}
