@charset "utf-8";

body {
  font-family: Verdana, sans-serif;
  color: #000;
  background: #fff;
}

.heading {
  font-weight: bold;
  font-size: 158%;
  text-align: center;
}

h1 {
  font-size: 150%;
  font-weight: bold;
}

h2 {
  font-size: 125%;
  font-weight: bold;
}

h3 {
  font-size: 100%;
  font-weight: bold;
  text-decoration: underline;
}

a:link, a:visited{
  color: #0067ab;
  text-decoration: none;
}

a:focus {
  outline:1px dotted;
}

a:hover, a:active {
  color: #0067ab;
  outline: none;
  text-decoration: underline;
}

.toc {
  margin: 0 20px 10px 20px;
  border: 1px solid #BBB;
  background-color: #dda;
  color: #000;
  background: #EEE;
  padding: 1px 10px;
}

.toc ul {
  list-style-type: none;
  padding-left: 0;
}

.content {
  margin: 0 20px 10px 20px;
  border: 1px solid #BBB;
  background-color: #dda;
  color: #000;
  background: #EEE;
  padding: 1px 10px;
}

.code {
  background-color: #ccc;
  font-family: monospace;
}

.classname, .command, .commandname, .constant, .filename, .function, .literal, .varname {
  font-family: monospace;
}

.option {
  color: #be5303;
  font-family: monospace;
}

.parameter {
  font-family: monospace;
  font-style: italic;
  border-bottom: 1px dashed black;
}

.productname {
  font-style: italic;
}

.important-title, .note-title {
  text-decoration: underline;
  font-weight: bold;
}

.important, .note {
  color: #000;
  background-color: #fae8b8;
  border: 1px solid #be5303;
  border-collapse: collapse;
  margin: 10px 30px;
  padding: 0 10px;
}

.cmdsynopsis, .programlisting {
  font-family: monospace;
  color: #000;
  background-color: #d0d0d0;
  border: 1px solid #0067ab;
  padding: 3px 5px;
  margin: 0 5px;
}

.screen {
  font-family: monospace;
  background-color: #d0d0d0;
  border: 1px solid #0067ab;
  padding: 3px 5px;
  margin: 0 5px;
}

img {
  border: none;
}

table.variablelist {
  color: #fff;
  background-color: #000;
  border: 1px solid #d0d0d0;
  border-collapse: collapse;
}

table.variablelist th {
  color: #000;
  background-color: #ccc;
}

table.variablelist td {
  color: #000;
  background-color: #fff;
  border: 1px solid #d0d0d0;
  padding: 3px;
  vertical-align: top;
}

table.variablelist td.varlistoption {
  white-space: nowrap;
}

td p:first-child {
  margin-top: 0;
}

td p:last-child {
  margin-bottom: 0;
}

a.headerlink {
  visibility: hidden;
  font-size: .8em;
  margin-left: .2em;
  opacity: .7;
  padding: 0 4px;
  text-decoration: none;
  transition: all .2s ease-out;
  user-select: none;
}

h1:hover > a.headerlink,
h2:hover > a.headerlink,
h3:hover > a.headerlink,
h4:hover > a.headerlink,
h5:hover > a.headerlink,
h6:hover > a.headerlink,
dt:hover > a.headerlink,
td:hover > a.headerlink,
caption:hover > a.headerlink,
p.caption:hover > a.headerlink,
div.code-block-caption:hover > a.headerlink {
  visibility: visible;
}
