#include "winuser.h"

mkvextract ICON  "../../share/icons/windows/mkvextract.ico"

1 RT_MANIFEST "manifest.xml"

1 VERSIONINFO
 FILEVERSION 99, 0, 0, 0
 PRODUCTVERSION 99, 0, 0, 0
 FILEFLAGSMASK 0x3f
 FILEOS 0x4
BEGIN
  BLOCK "StringFileInfo"
  BEGIN
    BLOCK "040904B0"
    BEGIN
      VALUE "CompanyName", "Moritz Bunkus"
      VALUE "FileDescription", "mkvextract"
      VALUE "FileVersion", "99.0"
      VALUE "InternalName", "mkvextract"
      VALUE "LegalCopyright", "Released under the GNU Pulic License http://www.gnu.org/licenses/gpl-2.0.html"
      VALUE "OriginalFilename", "mkvextract"
      VALUE "ProductName", "MKVToolNix"
      VALUE "ProductVersion", "99.0"
    END
  END
  BLOCK "VarFileInfo"
  BEGIN
    VALUE "Translation", 0x0409, 1252 // Englisch (USA)
  END
END
