#include "common/common_pch.h"

#include "mkvtoolnix-gui/mime_types.h"

namespace mtx::gui {

QString const MimeTypes::MergeSourceFileModelItem{"application/x-mkvtoolnixgui-sourcefilemodelitems"};

}
